--[[
	Todo:
	Only dim in combat.
]]

--~ local tostring = tostring
--~ 
--~ local LibStub = LibStub
--~ local pairs	= pairs
--~ local _G = _G

local folder, core = ...
_G["_DA"] = core

local curseVersion = GetAddOnMetadata(folder, "X-Curse-Packaged-Version")
local curseRevision = GetAddOnMetadata(folder, "X-Curse-Packaged-Revision") -- 133
if curseVersion and curseRevision and curseVersion ~= "r"..curseRevision then
	curseVersion = curseVersion.." (r"..curseRevision..")" --alpha versions will show "r99", tagged versions will show "1.2.3beta (r99)"
end

core.title		= GetAddOnMetadata(folder, "Title")
core.version	= curseVersion or "[dev]"
core.titleFull	= core.title.." "..(curseVersion and "v"..tostring(core.version) or tostring(core.version))
core.addonDir = "Interface\\AddOns\\"..folder.."\\"

LibStub("AceAddon-3.0"):NewAddon(core, folder, "AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0")

local L = LibStub("AceLocale-3.0"):GetLocale(folder, true)

local regEvents = {
	"ACTIONBAR_SLOT_CHANGED",
	"PLAYER_REGEN_DISABLED",
	"PLAYER_REGEN_ENABLED",
--~ 	"UPDATE_BONUS_ACTIONBAR",
}

core.defaultSettings = {
--~ 	global = {
--~ 		playerInfo = {},
--~ 		playerGems = {},
--~ 	},
--~ 	factionrealm = {
--~ 		knownCuts = {},
--~ 		myInventory = {},
--~ 		myAuctions = {},
--~ 	},
	profile = {}
}
 
core.stackableSpells = {
	[33745] = 3, --Lacerate (Druid)
	[33763] = 3, --Lifebloom (Druid)
	[7386] = 3, --Sunder Armor (Warrior)
	[770] = 3, --Faerie Fire (Druid)
	[16857] = 3, --Faerie Fire (Feral) (druid)
	[30451] = 4, --Arcane Blast (Mage)
	[31801] = 5, --Seal of Truth (Paladin)
	[74001] = 5, --Combat Readiness (Rogue)
	[80398] = 3, --Dark Intent (Warlock)
}

local blizBars = {
	"MultiBarBottomRight",
	"MultiBarBottomLeft",
	"MultiBarRight",
	"MultiBarLeft",
	"BonusAction",
	"Action",
}
 
local tostring = tostring
local chatFrame = _G["ChatFrame1"]

local DEBUG = false
--[===[@debug@
DEBUG = true
--@end-debug@]===]
local strWhiteBar		= "|cffffff00 || |r" -- a white bar to seperate the debug info.
local colouredName		= "|cff7f7f7f{|r|cffff0000DA|r|cff7f7f7f}|r "
local function echo(...)
	local tbl  = {...}
	local msg = tostring(tbl[1])
	for i=2,#tbl do 
		msg = msg..strWhiteBar..tostring(tbl[i])
	end

	local cf = chatFrame
	if cf then
		cf:AddMessage(colouredName..msg,.7,.7,.7)
	end
end
core.echo = echo

local strDebugFrom		= "|cffffff00[%s]|r" --Yellow function name. help pinpoint where the debug msg is from.
-----------------------------
local function Debug(from, ...)	--
-- simple print function.	--
------------------------------
--~ 	if DEBUG == false then
--~ 		return 
--~ 	end
	if core.db.profile.debugMessages == false then
		return
	end
	local tbl  = {...}
	local msg = tostring(tbl[1])
	for i=2,#tbl do 
		msg = msg..strWhiteBar..tostring(tbl[i])
	end
	echo(strDebugFrom:format(from).." "..msg)
end
core.Debug = Debug

local coreOpts

do
	local LibStub = LibStub
	function core:OnInitialize()
		self.db = LibStub("AceDB-3.0"):New("DA2_DB", core.defaultSettings, true) --'Default'
	
		self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
		self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
		self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")
		self.db.RegisterCallback(self, "OnProfileDeleted", "OnProfileChanged")
	
	--~ 	if db.global.lastVersion ~= self.version then
	--~ 		db.global.lastVersion = self.version
	--~ 		StaticPopup_Show(folder.."VersionChange", self.version);
	--~ 	end
		self:Hook("SecureHandlerSetFrameRef", true, nil, true)
	end
end

core.handerFrames = {}

----------------------------------------------------------------------------------------------------------
function core:SecureHandlerSetFrameRef(frame, label, refFrame)											--
-- Macrotexter and another addon I'm working on uses this function.										--
-- I don't know where else to get a frame's header frame so I'm hooking this function and saving it.	--
----------------------------------------------------------------------------------------------------------
	Debug("SecureHandlerSetFrameRef", "frame:"..tostring(frame and frame:GetName()), "label:"..tostring(label), "refFrame:"..tostring(refFrame and refFrame:GetName()))
	self.handerFrames[frame] = {
		label=label, 
		refFrame = refFrame
	}
end

----------------------------------------------------------------------
function core:OnProfileChanged(...)									--
-- User has reset proflie, so we reset our spell exists options.	--
----------------------------------------------------------------------
--~ 	-- Shut down anything left from previous settings
	self:Disable()
--~ 	-- Enable again with the new settings
	self:Enable()
end

do
	local pairs = pairs
	local _G = _G
	function core:OnEnable()
		for i, event in pairs(regEvents) do 
			self:RegisterEvent(event)
		end
	
		self:CollectActionInfo()
		
		if self.db.profile.dimCombatOnly == false then
			self:HookActionButtons()
		end
		
		if self.db.profile.hideBarBackground == true then
			self:HideBackgrounds(0)
		else
			self:HideBackgrounds(1)
		end
		
		if _G.Bartender4 or _G.Dominos then
			echo("********************************************")
			echo(L.noBartenderDominosSupport)
			echo("********************************************")
		end
	--~ 	Debug("OnEnable")
	end
end

function core:OnDisable()
	self:UnhookAllBars()
end


do
	local NUM_ACTIONBAR_BUTTONS = NUM_ACTIONBAR_BUTTONS
	local pairs = pairs
	local _G = _G
	

	--------------------------------------
	function core:HookActionButtons()		--
	-- Hook action button's OnUpdate.	--
	--------------------------------------
--~ 		Debug("HookActionButtons")
		for _, barName in pairs(blizBars) do 
			for i=1, NUM_ACTIONBAR_BUTTONS, 1 do
				self:HookActionButton(_G[barName.."Button"..i], barName, i)
			end
		end
		
	end
end

do
	local _G = _G
	--~ /run MainMenuBarArtFrame:SetAlpha(1)
	function core:HideBackgrounds(alpha)--	/run _DA:HideBackgrounds(0)
	--~ 	Debug("HideBackgrounds", alpha)
		--/run _G["BonusActionBarFrameTexture1"]:SetVertexColor(1,1,1,0)
		for i=1, 2 do 
			_G["BonusActionBarFrameTexture"..i]:SetVertexColor(1,1,1,alpha)
	--~ 		_G["ActionBarFrameTexture"..i]:SetVertexColor(1,1,1,alpha)
			
	--~ 		_G["MainMenuBarTexture"..i]:SetVertexColor(1,1,1,alpha)
	--~ 		_G["MainMenuBarTexture"..i]:SetVertexColor(1,1,1,alpha)
		end
		
		for i=0, 1 do 
			_G["MainMenuBarTexture"..i]:SetVertexColor(1,1,1,alpha)
		end
	end
end

do
	local _G = _G
	--Grab the button's iron below a bonus action button.
	local underButtonName, underButton
	core.underIcon = setmetatable({}, {
		__index = function(t,button)
			local underButtonName = "ActionButton"..button:GetID()
			local underButton = _G[underButtonName]
			t[button] = _G[underButtonName.."Icon"] or underButton:GetNormalTexture()
--~ 			print("underIcon", button, button:GetName(), underButton:GetName())
			return t[button]
		end
	})
end

do 
	local red, green, blue, alpha, ours
	----------------------------------------------------------------------------------------------------------------------
	function core:SetVertexColor(frame, ...)																			--
	-- Our SetVertexColor hook. When we call it we add a 5th var to varifiy it's our call. We block all other calls.	--
	-- This is done so that Bartender's built in range/mana colouring doesn't fight with us.							--
	----------------------------------------------------------------------------------------------------------------------
		red, green, blue, alpha, ours = ...
		if ours == true then
			core.hooks[frame].SetVertexColor(frame, ...)
		end
	end
end

do
	local _G = _G
	------------------------------------------------------
	function core:HookVertexColour(button)					--
	-- We hook the SetVertexColor function so we can	--
	-- block other addons from colouring the actions.	--
	-- This is optional.								--
	------------------------------------------------------
		local name = button:GetName()
		local icon = _G[name.."Icon"]
		if icon and not core:IsHooked(icon, "SetVertexColor") then
			core:RawHook(icon, "SetVertexColor", nil, true)
		end
	end
end

do 
	local string_find = string.find
	local strsplit = strsplit
	local tonumber = tonumber
	local function ItemLinkToId(itemLink) -- Itemlink, returns the itemID
		local _, _, itemString = string_find(itemLink, "^|c%x+|H(.+)|h%[.+%]")
		if itemString then
			local _, itemId, _, _, _, _, _, _, _ = strsplit(":", itemString)
			return tonumber(itemId)
		end
		return nil
	end
	
	core.itemLinkToID = setmetatable({}, {
		__index = function(t,i)
--~ 		tooltip:SetHyperlink("spell:"..i)
--~ 		tooltip:Show()
--~ 		lines = tooltip:NumLines()
			t[i] = ItemLinkToId(i) or 0
			return t[i]
		end
	})
	
end




do
	local IsAltKeyDown = IsAltKeyDown
	local UnitExists = UnitExists
	local IsUsableAction = IsUsableAction
	local ActionHasRange = ActionHasRange
	local IsActionInRange = IsActionInRange
	local GetActionInfo = GetActionInfo
	local GetSpellBookItemName = GetSpellBookItemName
	local BOOKTYPE_SPELL = BOOKTYPE_SPELL
	local GetSpellCooldown = GetSpellCooldown
	local GetMacroItem = GetMacroItem
	local GetMacroSpell = GetMacroSpell
	local _G = _G
	local GetBonusBarOffset = GetBonusBarOffset
	
	local spellName
--~ 	local underIcon
	------------------------------------------------------------------------------------------
	function core:UpdateActionButton(button, x)														--
	-- Our main OnUpdate function, here we check if we should show or dim a action button.	--
	------------------------------------------------------------------------------------------
		local name = button:GetName()
		local icon = _G[name.."Icon"] or button:GetNormalTexture()
		
--~ 		print("UpdateActionButton", button, name)
		if name:find("BonusAction") then
			self.underIcon[button]:SetVertexColor(1,1,1, 0, true)
		elseif name:find("ActionButton") and GetBonusBarOffset() > 0 then
			icon:SetVertexColor(1,1,1, 0, true)
			return
		end

		if self.db.profile.showAlt == true and IsAltKeyDown() then
--~ 			self:SetAlpha(alpha)
			icon:SetVertexColor(1, 1, 1, 1, true)
			return
		end
		
		local targetID = button:GetAttribute("unit1") or button:GetAttribute("unit") or "target"
--~ 		Debug("UpdateActionButton 1",name, button:GetAttribute("unit"), button:GetAttribute("type"),  button:GetAttribute("unitsuffix"), "hi")
		if self.db.profile.showNoTarget == true and not UnitExists(targetID) then
--~ 			self:SetAlpha(1)
			icon:SetVertexColor(1, 1, 1, 1, true)
			return
		end
		
		
		local actionID = button.action
		local hasTarget = true
		if actionID then
			local actionType, spellID, subType = GetActionInfo(actionID)
			--[[]]
			local buttonType = button:GetAttribute("type")
			local actionType, spellID, subType = GetActionInfo(actionID)
			local spellName
			local nTargetID
			
--~ 			if buttonType == "click" then
--~ 				realButton = button:GetAttribute("clickbutton") --The real button this button is clicking on.
--~ 				
--~ 				local header = handerFrames[realButton] and handerFrames[realButton].refFrame --self:GetFrameRef("SA_headerFrame")
			
--~ 			else
			if actionType == "macro" then --and self.db.profile.dimMacros == true then
			
			
			
			
				if self.db.profile.dimMacros == true then
					local nTargetID = self:GetMacroTarget(spellID) --Item's don't default to having a target. user needs to have [@target] in the /use macro.
					local item, itemLink = GetMacroItem(spellID)
					spellName = GetMacroSpell(spellID)
		
					if (item or spellName) and not nTargetID then
						hasTarget = false
					end
		
					targetID = nTargetID or targetID
					if item then
						actionType = "item"
						spellID = self.ItemLinkToId[itemLink]
						
					elseif spellName and self.spellNames[spellName] then
						actionType, subType = "spell", "spell"
						spellID = self.spellNames[spellName].sID--Changing spellID to a real spellID. currently it's a macroID.
					elseif not spellName then
						if self.db.profile.showNoTarget == true then
--~ 							self:SetAlpha(1)
							icon:SetVertexColor(1, 1, 1, 1, true)
							return
						else
--~ 							self:SetAlpha(self.db.profile.colours.unusable.a)
							icon:SetVertexColor(self.db.profile.colours.unusable.r, self.db.profile.colours.unusable.g, self.db.profile.colours.unusable.b, self.db.profile.dimAlpha, true)
							return
						end
					end
				else
--~ 					self:SetAlpha(1)
					icon:SetVertexColor(1, 1, 1, 1, true)
					return
				end
			
--~ 			elseif actionType == "macro" and self.db.profile.dimMacros == true then
			end		
			


		
			local isUsable, notEnoughMana = IsUsableAction(actionID)
			if self.db.profile.dimUnuseableOntarget == true and not isUsable and not notEnoughMana then --again with unusable this time using IsUsableAction() which is affected when a spell is needed inorder to cast a spell, (judgment needing seal)
--~ 				self:SetAlpha(self.db.profile.colours.unusable.a)
				icon:SetVertexColor(self.db.profile.colours.unusable.r, self.db.profile.colours.unusable.g, self.db.profile.colours.unusable.b, self.db.profile.dimAlpha, true)
				return
			end
			
			if self.db.profile.dimNoMana == true and notEnoughMana then
--~ 				self:SetAlpha(self.db.profile.colours.mana.a)
				icon:SetVertexColor(self.db.profile.colours.mana.r, self.db.profile.colours.mana.g, self.db.profile.colours.mana.b, self.db.profile.dimAlpha, true)
				return
			end
			
			if ActionHasRange(actionID) then
				if self.db.profile.dimUnuseableOntarget == true and IsActionInRange(actionID,targetID) == nil then 
				
					--Testing code that hides a spell if it's harm and target is friendly, or a help spell and target's hostile. There were problems...
	--~ 				if actionType and actionType == "spell" and subType  == "spell" and spellID and spellID > 0 then
	--~ 					if UnitExists(targetID) and spellID and ((IsHarmfulSpell(spellID, BOOKTYPE_SPELL) and UnitReaction(targetID, "player") >= 5) or (IsHelpfulSpell(spellID, BOOKTYPE_SPELL) and UnitReaction(targetID, "player") <= 4)) then
	--~ 						self:SetAlpha(self.db.profile.colours.unusable.a)
	--~ 						icon:SetVertexColor(self.db.profile.colours.unusable.r, self.db.profile.colours.unusable.g, self.db.profile.colours.unusable.b, nil, true)
	--~ 						return
	--~ 					end
	--~ 				end
					
					if not isUsable and not notEnoughMana then
--~ 						self:SetAlpha(self.db.profile.colours.unusable.a)
						icon:SetVertexColor(self.db.profile.colours.unusable.r, self.db.profile.colours.unusable.g, self.db.profile.colours.unusable.b, self.db.profile.dimAlpha, true)
						return
					end
				end
				if self.db.profile.dimOutOfRange == true and IsActionInRange(actionID,targetID) == 0 then
--~ 					self:SetAlpha(self.db.profile.colours.range.a)
					icon:SetVertexColor(self.db.profile.colours.range.r, self.db.profile.colours.range.g, self.db.profile.colours.range.b, self.db.profile.dimAlpha, true)
					return
				end
			end
			
			if (actionType == "spell" and subType == "spell" and spellID and spellID > 0) then
				spellName = GetSpellBookItemName( self.sbID[spellID], BOOKTYPE_SPELL ) --spellName or 
				
				
				if self.db.profile.dimCooldown == true then
					local start, duration, enabled = GetSpellCooldown(spellName)
					if start and start > 0 then --and ( start > 0 and duration > 0) then
						if start ~= self:GetGCD() then
	--~ 						local wait = start + duration - GetTime_()
	--~ 						if wait > 1 then--If there's less then 1 second remaining in the cooldown, show the action. This gives user a second warning to deside if they'll start mashing it.
								icon:SetVertexColor(self.db.profile.colours.cooldown.r, self.db.profile.colours.cooldown.g, self.db.profile.colours.cooldown.b, self.db.profile.dimAlpha, true)
								return
	--~ 						end
						end
					end
				end
	
				if self.db.profile.dimSpellExists == true and UnitExists(targetID) then
--~ 					if self.db.profile.se[spellName] == true then
						if self:IsBuffOn(spellName, targetID, spellID) then
							icon:SetVertexColor(self.db.profile.colours.exists.r, self.db.profile.colours.exists.g, self.db.profile.colours.exists.b, self.db.profile.dimAlpha, true)
							return
						end
--~ 					end
				end
				
				
			end
			
			
		end
		
		icon:SetVertexColor(1, 1, 1, 1, true)
	end
end

core.hookedButtons = {}

local ActionButton_OnUpdate
do
	local lastUpdated = {}
--~ 	local updateFrequency = 0.1
	--------------------------------------------------
	function ActionButton_OnUpdate(self, elapsed)	--
	-- Our OnUpdate hooked function. 				--
	--------------------------------------------------
	--~ 	self.daLastUpdate = self.daLastUpdate + elapsed
	--~ 	if self.daLastUpdate >= self.db.profile.updateFrequency then 
	--~ 		self.daLastUpdate = 0
		lastUpdated[self] = (lastUpdated[self] or 0) + elapsed
		
		if lastUpdated[self] >= core.db.profile.updateFrequency then 
			lastUpdated[self] = 0
			core:UpdateActionButton(self)
		end
	--~ 	end
	end
end

do
	
	--------------------------------------------------
	function core:HookActionButton(buttonFrame, bar, id)	--
	-- Hook a action button's OnUpdate script.		--
	--------------------------------------------------
	--~ 	Debug("HookActionButton 1", buttonFrame, bar, id)
		if buttonFrame then
			self.hookedButtons[bar] = self.hookedButtons[bar] or {}
			if id == 0 then
				id = #self.hookedButtons[bar] + 1
			end
			self.hookedButtons[bar][id] = buttonFrame
	--~ 		Debug("HookActionButton 2", buttonFrame, bar, id)
			
			if buttonFrame:HasScript("OnUpdate") and not core:IsHooked(buttonFrame, "OnUpdate") then
				buttonFrame.daLastUpdate = 0
				core:HookScript(buttonFrame, "OnUpdate", ActionButton_OnUpdate)
	--~ 			Debug("HookActionButton 3", buttonFrame, bar, id)
			end
			
			if self.db.profile.blockOtherColours == true then
				self:HookVertexColour(buttonFrame)
			end
		end
	end
end



do
	local GetMacroBody = GetMacroBody
	local SecureCmdOptionParse = SecureCmdOptionParse
	
	local body
	local _, target
	function core:GetMacroTarget(macroID)
		body = GetMacroBody(macroID)
		if body then
			_, target = SecureCmdOptionParse(body)
			return target
		end
		return nil
	end
end


do
	local GetMouseFocus = GetMouseFocus
	local ActionHasRange = ActionHasRange
	local IsActionInRange = IsActionInRange
	local GetActionInfo = GetActionInfo
	local GetMacroBody = GetMacroBody
	local SecureCmdOptionParse = SecureCmdOptionParse
	local GetMacroSpell = GetMacroSpell
	local GetMacroItem = GetMacroItem
	
	------------------------------
	function core:GetMouseoverInfo()	--	/script _DA:GetMouseoverInfo()
	-- Used to debug.			--
	------------------------------
		local self = GetMouseFocus()--	/script print(GetMouseFocus():GetName().." - "..GetMouseFocus():GetParent():GetName())
		
		local targetID = "target"
		
		Debug("GetMouseoverInfo", self:GetName(), GetMouseFocus():GetParent(), self.action)
			if self.action then
			local actionID = self.action
		--~ 	name, iconTexture, body, local = GetMacroInfo(MacroID)
		--~ 		local targetID = "target"
			Debug("GetMouseoverInfo", ActionHasRange(actionID), IsActionInRange(actionID,targetID))
		
			local actionType, spellID = GetActionInfo(actionID)
			Debug("GetMouseoverInfo 1", actionType, spellID)
		--~ 		local spellName, spellRank = GetSpellBookItemName( spellID, BOOKTYPE_SPELL )
			if actionType == "macro" then
				local body = GetMacroBody(spellID)
				
				Debug("Body", body)
				
				Debug("SecureCmdOptionParse", SecureCmdOptionParse(body))
		
				Debug("GetMacroSpell", GetMacroSpell(spellID))
				Debug("GetMacroTarget", self:GetMacroTarget(spellID))
				Debug("GetMacroItem", GetMacroItem(spellID))
				
			end
		end
	end
end

do
	local actionID
	function core:CollectActionInfo()
		
		for b=0, 9 do
			for s = 1, 12 do
				actionID = s+(b*12)
				self:SaveActionInfo(actionID)
			end
		end
		self:CollectSpellInfo()
	--~ 	if self.db.profile.dimSpellExists == true then
	--~ 		AddSpellExistOptions()
	--~ 	end
		
	end
end

do
	local GetSpellBookItemName = GetSpellBookItemName
	local BOOKTYPE_SPELL = BOOKTYPE_SPELL
	local IsPassiveSpell = IsPassiveSpell
	
	------------------------------------------------------------------------------
	function core:CollectSpellInfo()													--	/script DA.CollectSpellInfo()
	-- Save spellID's to a table indexed by spellName. This is used for macros.	--
	-- So we know the spellID for any spells within a macro.					--
	------------------------------------------------------------------------------
		local spellID = 1
		
		while( GetSpellBookItemName(spellID, BOOKTYPE_SPELL) ) do --find the top rank of every spell
			local spellName, spellRank = GetSpellBookItemName( spellID, BOOKTYPE_SPELL )
	--~ 		local nextSpell = GetSpellBookItemName( spellID+1, BOOKTYPE_SPELL )
			
			if not IsPassiveSpell(spellID, BOOKTYPE_SPELL) and not self.spellNames[spellName] then
--~ 				local targetSpell, canStack = GetSpellDetails(spellName, spellRank)
				self.spellNames[spellName] = {sID=spellID} --, targetSpell=targetSpell, canStack = canStack}
			end
			spellID = spellID + 1
		end
	end
end

core.spellNames = {}

do
	local HasAction = HasAction
	local GetActionInfo = GetActionInfo
	local GetSpellBookItemName = GetSpellBookItemName
	local BOOKTYPE_SPELL = BOOKTYPE_SPELL
	local FindSpellBookSlotBySpellID = FindSpellBookSlotBySpellID
	
	local actionType, spellID, subType, spellRank, spellName, sID2
	------------------------------------------------------------------------------------------------------
	function core:SaveActionInfo(actionID)																	--
	-- Save a actionID's name, spellID, weather it can be cast on a target and if it has a cooldown.	--
	------------------------------------------------------------------------------------------------------
		
		if HasAction(actionID) then -- and ActionHasRange_(actionID)
			actionType, spellID, subType  = GetActionInfo(actionID)
			
			if actionType == "spell" and subType  == "spell" and spellID and spellID > 0 then
				
				local spellName, spellReq = GetSpellBookItemName( self.sbID[spellID], BOOKTYPE_SPELL )
			
--~ 				Debug("SaveActionInfo 1", actionID, actionType, spellID, subType, spellName, spellRank )
				if spellName then
					self.spellNames[spellName] = {aID=actionID, sID=spellID, sbID=FindSpellBookSlotBySpellID(spellID)}
				end
			end
		end
	end
end

local targetSpells = {--spells that can be cast on other people.
	["the target"] = true,
	["the enemy"] = true,
	["enemy target"] = true,
	["friendly target"] = true,
	["friendly party"] = true,
}

--[[
do
	local GetSpellLink = GetSpellLink
	local SpellHasRange = SpellHasRange
	local FindSpellBookSlotBySpellID = FindSpellBookSlotBySpellID
	local BOOKTYPE_SPELL = BOOKTYPE_SPELL
	local tonumber = tonumber
	
	
	function core:GetSpellDetails(spellID)	
		local link = GetSpellLink(spellID)
		Debug("GetSpellDetails 1", spellID, link)
		
		
		local canStack = 0
		local otherSpell = false
		if SpellHasRange(spellID) then
			local description = self.spellDescription[spellID]
			if description ~= "" then
				if description:find("stack") or description:find("applied") then	--LOCALE
					for times in description:gmatch("up to (.+) times") do
						canStack = tonumber(times or 0)
						break
					end
				end
			
				for tText in pairs(targetSpells) do 

					if (description:find("over") or description:find("every") or description:find("sec") or description:find("min")) and description:find(tText) then
						otherSpell = true
					end
				end
			end
		end
		
		return canStack, otherSpell
		
--~ 		tooltip:SetHyperlink("spell:"..spellID)
--~ 		tooltip:Show()
		
--~ 		local lines = tooltip:NumLines()
		
--~ 		Debug("GetSpellDetails 2", lines)
--~ 		if lines > 0 then
--~ 			local description = (_G[folder.."ToolTipTextLeft"..lines]:GetText() or "")
--~ 			Debug("GetSpellDetails 3", description)
--~ 			Debug("GetSpellDetails 4", SpellHasRange(spellID), SpellHasRange(FindSpellBookSlotBySpellID(spellID), BOOKTYPE_SPELL))
			
			if SpellHasRange(spellName) then
				for tText in pairs(targetSpells) do 
					if description:find(L["stack"]:lower()) or description:find(L["applied"]:lower()) then
						for times in description:gmatch(L["up to (.+) times"]) do
							canStack = tonumber(times or 0)
							break
						end
					end
					
					if (description:find(L["over"]:lower()) or description:find(L["every"]:lower()) or description:find(L["sec"]:lower()) or description:find(L["min"]:lower())) and description:find(tText) then
						otherSpell = true
					end
				end
				
				
			end
			
--~ 		end
	end

end]]

do
	local _G = _G
	
	local tooltip = CreateFrame("GameTooltip", folder.."ToolTip", UIParent, "GameTooltipTemplate")
	--~ local tooltipStats = _G[tooltip:GetName().."TextLeft2"]
	tooltip:SetOwner(UIParent, "ANCHOR_NONE")

	local lines
	core.spellDescription = setmetatable({}, {
		__index = function(t,i)
		tooltip:SetHyperlink("spell:"..i)
		tooltip:Show()
		lines = tooltip:NumLines()
		t[i] = _G[folder.."ToolTipTextLeft"..lines]:GetText() or ""
		
		return t[i]
		end
	})
end

do
	local FindSpellBookSlotBySpellID = FindSpellBookSlotBySpellID
	core.sbID = setmetatable({}, {
		__index = function(t,i)
		t[i] = FindSpellBookSlotBySpellID(i)
	--~ 	Debug("canCutGem", iLink[i], t[i])
		return t[i]
		end
	})
end

--[[
do
	local BOOKTYPE_SPELL = BOOKTYPE_SPELL
	local SpellHasRange = SpellHasRange
	core.spellHasRange = setmetatable({}, {
		__index = function(t,i)
		t[i] = SpellHasRange( core.sbID[i] , BOOKTYPE_SPELL)
	--~ 	Debug("canCutGem", iLink[i], t[i])
		return t[i]
		end
	})
end
]]

do
	--------------------------------------------------------------
	function core:PLAYER_REGEN_DISABLED(event, ...)			--
	--------------------------------------------------------------
		if self.db.profile.dimCombatOnly == true then
			self:HookActionButtons()
		end
	end
	--------------------------------------------------------------
	function core:PLAYER_REGEN_ENABLED(event, ...)			--
	--------------------------------------------------------------
		if self.db.profile.dimCombatOnly == true then
			self:UnhookAllBars()
		end
	end
end


--[[
do
	local GetBonusBarOffset = GetBonusBarOffset
	local _G = _G
	
	local texture;
	--------------------------------------------------------------
	function core:UPDATE_BONUS_ACTIONBAR(event, ...)			--
	--------------------------------------------------------------
		if ( GetBonusBarOffset() > 0 ) then
			for s = 1, 12 do
				texture = _G["ActionButton"..s.."Icon"] or _G["ActionButton"..s]:GetNormalTexture()
--~ 				Debug(event, s)
				texture:SetVertexColor(0,0,0, 0, true)
			end
		end
	end
end
]]

do
	local pairs = pairs
	----------------------------------------------------------------------
	function core:UnhookAllBars()											--
	-- Unhook all bars, this also returns action buttons to 100% alpha.	--
	----------------------------------------------------------------------
--~ 		Debug("UnhookAllBars")
	--~ 	if BT4 == true then
	--~ 		for i=1, 10 do 
	--~ 			UnhookBar("BT4Bar"..i)
	--~ 		end
	--~ 		
	--~ 	elseif Dominos == true then
	--~ 		for i=1, 10 do 
	--~ 			UnhookBar("dBar"..i)
	--~ 		end
	--~ 		
	--~ 	else
	--~ 		self:UnhookBar("Bar1")
	--~ 		self:UnhookBar("BonusBar1")
	--~ 		self:UnhookBar("MultiBarBottomLeft")
	--~ 		self:UnhookBar("MultiBarBottomRight")
	--~ 		self:UnhookBar("MultiBarRight")
	--~ 		self:UnhookBar("MultiBarLeft")
			
		for _, barName in pairs(blizBars) do 
			self:UnhookBar(barName)
		end
	--~ 	end
	end
end

do
	local pairs = pairs
	------------------------------------------
	function core:UnhookBar(bar)					--
	-- Onhook all actions used by a bar.	--
	------------------------------------------
		if self.hookedButtons[bar] then
	--~ 		local icon
			for _, buttonFrame in pairs(self.hookedButtons[bar]) do 
				if core:IsHooked(buttonFrame, "OnUpdate") then
					core:Unhook(buttonFrame, "OnUpdate")
				end
				
				--reset colour
				self:UnhookVertexColour(buttonFrame)
				buttonFrame:SetAlpha(1)
			end
		end
	end
end


do
	local _G = _G
	------------------------------------------------------
	function core:UnhookVertexColour(button)					--
	-- Release our hook on SetVertexColor()				--
	------------------------------------------------------
		local name = button:GetName()
		local icon = _G[name.."Icon"] or button:GetNormalTexture()
		if icon then
			if core:IsHooked(icon, "SetVertexColor") then
				core:Unhook(icon, "SetVertexColor")
			end
			icon:SetVertexColor(1, 1, 1, 1, true)
		end
	end
end

do
	local HasAction = HasAction
	local GetActionInfo = GetActionInfo
	local GetSpellBookItemName = GetSpellBookItemName
	local BOOKTYPE_SPELL = BOOKTYPE_SPELL
	local pairs = pairs
	
	local actionType, spellID, subType, spellName
	--------------------------------------------------------------
	function core:ACTIONBAR_SLOT_CHANGED(event, ...)			--
	-- A slot has changed, we save the spell's new actionID.	--
	--------------------------------------------------------------
		local actionID  = ...-- ***
		if actionID == 0 then--no actionID, save all actionIDs.
			self:CollectActionInfo()
			return
		end
		
		for spellName, data in pairs(self.spellNames) do 
			if data.aID == actionID then
				self.spellNames[spellName].aID = nil
				break
			end
		end
		if HasAction(actionID) then -- and ActionHasRange_(actionID)
			actionType, spellID, subType  = GetActionInfo(actionID)
			if actionType == "spell" and subType  == "spell" and spellID and spellID > 0 then
				spellName = GetSpellBookItemName( self.sbID[spellID], BOOKTYPE_SPELL )
				if spellName and self.spellNames[spellName] then
					self.spellNames[spellName].aID = actionID
--~ 					Debug(event,spellName.." is now aID: "..actionID)
				end
			end
		end
	end
end

do
	local UnitClass = UnitClass
	local function GetGCDName()
		local  _, class  = UnitClass('player')
		if class == 'DEATHKNIGHT' then
			return "Blood Strike"
		elseif class == 'DRUID' then
			return "Healing Touch"
		elseif class == 'HUNTER' then
			return "Serpent Sting"
		elseif class == 'MAGE' then
			return "Fireball"
		elseif class == 'PALADIN' then
			return "Flash of Light"
		elseif class == 'PRIEST' then
			return "Lesser Heal"
		elseif class == 'ROGUE' then
			return "Sinister Strike"
		elseif class == 'SHAMAN' then
			return "Healing Wave"
		elseif class == 'WARLOCK' then
			return "Immolate"
		elseif class == 'WARRIOR' then
			return 'Sunder Armor'
		end
	end
	local gcdName = GetGCDName()
	
	local GetSpellCooldown = GetSpellCooldown
	--------------------------------------------------------------------------
	function core:GetGCD()														--
	-- Get the global cooldown from a spell that normally has no cooldown.	--
	--------------------------------------------------------------------------
		local start = GetSpellCooldown(gcdName) --, duration,  enable
	--~ 	if enable ==  1 and start and duration > 0 and  duration <= 1.5 then
			return start or 0
	--~ 	end
	--~ 	return 0
	end
end

do
	local UnitDebuff = UnitDebuff
	local UnitBuff = UnitBuff
	
	function core:IsBuffOn(spellName, unitID, spellID)			--	/script DA.IsBuffOn("Lifebloom", "target")
		local name, _, _, count, _, _, _, unitCaster = UnitDebuff(unitID, spellName)
		if name then
			if unitCaster and unitCaster == "player" then
--~ 				if self.spellNames[spellName].canStack == 0 or count >= self.db.profile.sc[spellName] then
				if not self.stackableSpells[spellID] or count >= self.stackableSpells[spellID] then
					return true
				end
			end
		end
		name, _, _, count, _, _, _, unitCaster = UnitBuff(unitID, spellName)
		if name then
			if unitCaster and unitCaster == "player" then
--~ 				if self.spellNames[spellName].canStack == 0 or count >= self.db.profile.sc[spellName] then
				if not self.stackableSpells[spellID] or count >= self.stackableSpells[spellID] then
					return true
				end
			end
		end
		return nil
	end
end