local folder, core = ...

local LibStub = LibStub
local Debug = core.Debug
local echo = core.echo

local L = core.L or LibStub("AceLocale-3.0"):GetLocale(folder, true)

local defaultSettings = core.defaultSettings
defaultSettings.profile.debugMessages			= false




defaultSettings.profile.hideBarBackground			= true
--~ defaultSettings.profile.dimAlpha			= 0.1

defaultSettings.profile.showAlt			= true
defaultSettings.profile.showNoTarget			= true
defaultSettings.profile.dimUnuseableOntarget	= true
defaultSettings.profile.dimNoMana				= false
defaultSettings.profile.dimUnuseableOntarget	= true
defaultSettings.profile.dimOutOfRange			= true
defaultSettings.profile.dimCooldown				= true
defaultSettings.profile.dimSpellExists			= true
defaultSettings.profile.updateFrequency			= 0.1
defaultSettings.profile.dimMacros				= true
defaultSettings.profile.dimCombatOnly			= true

defaultSettings.profile.colours = {}
defaultSettings.profile.colours.range			= {r=1,g=0,b=0}	--red, 25% alpha
defaultSettings.profile.colours.unusable		= {r=1,g=1,b=1}	--white, 25%
defaultSettings.profile.colours.mana			= {r=0,g=0,b=1}	--mana, 25%
defaultSettings.profile.colours.cooldown		= {r=1,g=1,b=1}	--white, 25%
defaultSettings.profile.colours.exists			= {r=0,g=1,b=0}	--green, 25%
defaultSettings.profile.dimAlpha			= 0.1
defaultSettings.profile.blockOtherColours		= true


local coreOpts
do 
	local OnInitialize = core.OnInitialize
	function core:OnInitialize(...)
		if OnInitialize then OnInitialize(self, ...) end
	
--~ 		local config = LibStub("AceConfig-3.0")
--~ 		local dialog = LibStub("AceConfigDialog-3.0")
--~ 		config:RegisterOptionsTable(self.title, self:GetCoreOptionsTable())
--~ 		coreOpts = dialog:AddToBlizOptions(self.title, self.titleFull)
		
		self:AddBlizzardOptions()
		self:CreateOptionsDisplay()
		
		self:RegisterChatCommand("da", "ChatCommand")
	end
end



----------------------------------------------
function core:ChatCommand(input)	--
----------------------------------------------
	core:OpenOptionsFrame()
end

local name = function(info) 
	local key = info[#info]
	return L[key.."Name"] and L[key.."Name"]:format(core.db.profile[key]) or key
end
local desc = function(info) 
	local key = info[#info]
	return L[key.."Desc"] and L[key.."Desc"]:format(core.db.profile[key]) or key
end

function core:OpenOptionsFrame()
	LibStub("AceConfigDialog-3.0"):Open(self.title)
end

do
	local GENERAL = GENERAL
	local ENABLE = ENABLE
	
	function core:CreateOptionsDisplay()
		local db = core.db
		local options = {
			type = "group",
			name = core.title,
	--~ 		get = function( k ) return db[k.arg] end,
	--~ 		set = function( k, v ) db[k.arg] = v; end,
			
			get = function(info)
				local key = info[#info]
				return db.profile[key]
			end,
			set = function(info, v)
				local key = info[#info] 
				db.profile[key] = v
				self:Disable()
				self:Enable()
			end,
			
			
			args = {},
			plugins = {},
		}
		
		options.plugins["profiles"] = { profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(db) }
	
		options.args.general = {
			type = "group",
			name = GENERAL,
			order = 1,
			disabled = function(info) 
				if info.type ~= "group" then
					return not core:IsEnabled()
				end
				return false
			end,
			args = {
			
				enable = {
					type = "toggle",	order	= 1,
					name	= ENABLE,
					desc	= desc, --L.enableDesc,
					set = function(info,val) 
						if val == true then
							core:Enable()
						else
							core:Disable()
						end
					
					end,
					get = function(info) return core:IsEnabled() end,
					disabled = false,
				},
				
				debugMessages = {
					type = "toggle", order	= 10,
					name = "debug messages",
					desc = "Show debug messages in chat",
				},
				
			}
		}
		
		options.args.dimConditionsGroup = {
			type = "group",
			name = name, desc = desc,
			order = 3,
			disabled = function() return not core:IsEnabled() end,
			args = {
--~ 				hideBarBackground = {
--~ 					type = "toggle", order	= 1,
--~ 					name = name, desc = desc,
--~ 				},


				updateFrequency = {
					name = name, desc = desc,
					type = "range",
					order	= 1,
					min		= 0.1,
					max		= 1,
					step	= 0.1,
				},
				
				showAlt = {
					type = "toggle", order	= 2,
					name = name, desc = desc,
				},
				showNoTarget = {
					type = "toggle", order	= 3,
					name = name, desc = desc,
				},
				dimUnuseableOntarget = {
					type = "toggle", order	= 4,
					name = name, desc = desc,
				},
				dimNoMana = {
					type = "toggle", order	= 5,
					name = name, desc = desc,
				},
				dimUnuseableOntarget = {
					type = "toggle", order	= 6,
					name = name, desc = desc,
				},
				dimOutOfRange = {
					type = "toggle", order	= 7,
					name = name, desc = desc,
				},
				dimCooldown = {
					type = "toggle", order	= 8,
					name = name, desc = desc,
				},
				dimSpellExists = {
					type = "toggle", order	= 9,
					name = name, desc = desc,
--~ 					disabled = true,
				},
				dimMacros = {
					type = "toggle", order	= 10,
					name = name, desc = desc,
--~ 					disabled = true,
				},
				dimCombatOnly = {
					type = "toggle", order	= 11,
					name = name, desc = desc,
--~ 					disabled = true,
				},
				
				
			}
		}
		
		options.args.coloursGroup = {
			type = "group",
			name = name, desc = desc,
			order = 3,
			disabled = function() return not core:IsEnabled() end,
			args = {
--~ 				hideBarBackground = {
--~ 					type = "toggle", order	= 1,
--~ 					name = name, desc = desc,
--~ 				},
				
				dimAlpha = {
					name = name, desc = desc,
					type = "range",
					order	= 1,
					min		= 0,
					max		= 1,
					step	= 0.1,
					isPercent = true,
--~ 					set = function(info,val) 
--~ 						P.barOffsetX = val
--~ 						core:ResetAllBarPoints()
--~ 					end,
		--~ 			get = function(info) return P.barOffsetX end
				},
				
				rangeColour = {
--~ 		;~ 			name = L["Range"],	order	= 10,
--~ 		;~ 			desc = L["Target is out of range of action"],
					
					name = name, desc = desc,
					order	= 10,
					type = "color",
--~ 					hasAlpha	= true,
					set = function(info,r,g,b,a) 
						self.db.profile.colours.range.r = r
						self.db.profile.colours.range.g = g
						self.db.profile.colours.range.b = b
						self.db.profile.colours.range.a = a
					end,
					get = function(info) return self.db.profile.colours.range.r, self.db.profile.colours.range.g, self.db.profile.colours.range.b, self.db.profile.colours.range.a end
				},
			
				unusableColour = {
--~ 		;~ 			name = L["Unusable"],	order	= 11,
--~ 		;~ 			desc = L["Action not unsable (ex. heals on enemy)"],
					name = name, desc = desc,
					order	= 11,
					type = "color",
--~ 					hasAlpha	= true,
					set = function(info,r,g,b,a) 
						self.db.profile.colours.unusable.r = r
						self.db.profile.colours.unusable.g = g
						self.db.profile.colours.unusable.b = b
						self.db.profile.colours.unusable.a = a
					end,
					get = function(info) return self.db.profile.colours.unusable.r, self.db.profile.colours.unusable.g, self.db.profile.colours.unusable.b, self.db.profile.colours.unusable.a end
				},
			
				manaColour = {
--~ 		;~ 			name = "Mana",	order	= 12,
--~ 		;~ 			desc = "Not enough mana/range/energy/runic power to cast spell",
					name = name, desc = desc,
					order	= 12,
					type = "color",
--~ 					hasAlpha	= true,
					set = function(info,r,g,b,a) 
						self.db.profile.colours.mana.r = r
						self.db.profile.colours.mana.g = g
						self.db.profile.colours.mana.b = b
						self.db.profile.colours.mana.a = a
					end,
					get = function(info) return self.db.profile.colours.mana.r, self.db.profile.colours.mana.g, self.db.profile.colours.mana.b, self.db.profile.colours.mana.a end
				},
				
				cooldownColour = {
--~ 		;~ 			name = L["Coolodwn"],	order	= 13,
--~ 		;~ 			desc = L["Spell is on cooldown"],
					name = name, desc = desc,
					order	= 13,
					type = "color",
--~ 					hasAlpha	= true,
					set = function(info,r,g,b,a) 
						self.db.profile.colours.cooldown.r = r
						self.db.profile.colours.cooldown.g = g
						self.db.profile.colours.cooldown.b = b
						self.db.profile.colours.cooldown.a = a
					end,
					get = function(info) return self.db.profile.colours.cooldown.r, self.db.profile.colours.cooldown.g, self.db.profile.colours.cooldown.b, self.db.profile.colours.cooldown.a end
				},
				
				existsColour = {
--~ 		;~ 			name = L["Exists"],	order	= 14,
--~ 		;~ 			desc = L["Spell already exists on target"],
					name = name, desc = desc,
					order	= 14,
					type = "color",
--~ 					hasAlpha	= true,
					set = function(info,r,g,b,a) 
						self.db.profile.colours.exists.r = r
						self.db.profile.colours.exists.g = g
						self.db.profile.colours.exists.b = b
						self.db.profile.colours.exists.a = a
					end,
					get = function(info) return self.db.profile.colours.exists.r, self.db.profile.colours.exists.g, self.db.profile.colours.exists.b, self.db.profile.colours.exists.a end
				},
				
				blockOtherColours = {
					type = "toggle", order	= 20,
					name = name, desc = desc,
--~ 					disabled = true,
				},
								
			}
		}
		
		
--~ 		showNoTarget
		
		options.args.blizzardBarsGroup = {
			type = "group",
			name = name, desc = desc,
			order = 3,
			disabled = function() return not core:IsEnabled() end,
			args = {
				hideBarBackground = {
					type = "toggle", order	= 1,
					name = name, desc = desc,
				},
			}
		}
		
		
		local config = LibStub("AceConfig-3.0")
	--~ 	local dialog = LibStub("AceConfigDialog-3.0")
		config:RegisterOptionsTable(core.title, options ) --
	--~ 	coreOpts = dialog:AddToBlizOptions(core.title, core.titleFull)
		
		LibStub("AceConfigDialog-3.0"):SetDefaultSize(core.title, 600, 400) --680
		
		self.options = options
		
		self:BuildAboutMenu()
	end
end


do
	local UnitName = UnitName
	local HideUIPanel = HideUIPanel
	local InterfaceOptionsFrame = InterfaceOptionsFrame
	local GameMenuFrame = GameMenuFrame
	local CreateFrame = CreateFrame
	local table_wipe = table.wipe
	local coreOpts
	function core:AddBlizzardOptions()
	-- I'm going to use AceConfigDialog's window to show my options. 
	-- But I still want my addon listed in Blizzard's options frame. 
	-- The only thing shown in the Blizzard options frame will be a button 
	-- to open AceConfigDialog's window.
	
		local blizOptions = {
			name = core.name,
			type = "group",
			args = {
			
				Desc = {
					type = "description",
					name = L.blizOptionsName:format(UnitName("player")),
					order = 1,
				},
				openOptionsFrame = {
					type = "execute",	order	= 2,
					name	= name, --L.openOptionsFrameName,
					desc	= desc, --L.openOptionsFrameDesc,
					func = function(info, v)
						HideUIPanel(InterfaceOptionsFrame)
						HideUIPanel(GameMenuFrame)
						
						--If I open the AceConfigDialog window now it will close once HideUIPanel fires. 
						-- So I'm making a frame to open it on the next frame refresh.
						local f = CreateFrame("Frame")
						f:SetScript("OnUpdate", function(this, elapsed) 
							core:OpenOptionsFrame()
							this:Hide()
							table_wipe(this)
						end)
					end,
				},
			}
		}
	
		local config = LibStub("AceConfig-3.0")
		local dialog = LibStub("AceConfigDialog-3.0")
		local blizName = core.title.."bliz"
		
		config:RegisterOptionsTable(blizName, blizOptions )
		coreOpts = dialog:AddToBlizOptions(blizName, core.titleFull)
	end
end

do
	local tostring = tostring
	local GetAddOnMetadata = GetAddOnMetadata
	local pairs = pairs
	function core:BuildAboutMenu()
		local options = self.options
		
		options.args.about = {
			type = "group",
			name = L.about,
			order = 99,
			args = {
			}
		}
		
		local fields = {"Author", "X-Category", "X-License", "X-Email", "Email", "eMail", "X-Website", "X-Credits", "X-Localizations", "X-Donate", "X-Bitcoin"}
		local haseditbox = {["X-Website"] = true, ["X-Email"] = true, ["X-Donate"] = true, ["Email"] = true, ["eMail"] = true, ["X-Bitcoin"] = true}
	
		local fNames = {
			["Author"] = L.author,
			["X-License"] = L.license,
			["X-Website"] = L.website,
			["X-Donate"] = L.donate,
			["X-Email"] = L.email,
			["X-Bitcoin"] = L.bitcoinAddress,
		}
		local yellow = "|cffffd100%s|r"
		
		
		
		options.args.about.args.title = {
			type = "description",
			name = yellow:format(L.title..": ")..core.title,
			order = 1,
		}
		options.args.about.args.version = {
			type = "description",
			name = yellow:format(L.version..": ")..core.version,
			order = 2,
		}
		options.args.about.args.notes = {
			type = "description",
			name = yellow:format(L.notes..": ")..tostring(GetAddOnMetadata(folder, "Notes")),
			order = 3,
		}
	
		for i,field in pairs(fields) do
			local val = GetAddOnMetadata(folder, field)
			if val then
				
				if haseditbox[field] then
					options.args.about.args[field] = {
						type = "input",
						name = fNames[field] or field,
						order = i+10,
						desc = L.clickCopy,
						width = "full",
						get = function(info)
							local key = info[#info]
							return GetAddOnMetadata(folder, key)
						end,	
					}
				else
					options.args.about.args[field] = {
						type = "description",
						name = yellow:format((fNames[field] or field)..": ")..val,
						width = "full",
						order = i+10,
					}
				end
		
			end
		end
	
		LibStub("AceConfig-3.0"):RegisterOptionsTable(core.title, options ) --
	end
end